%%This is a part of multi-layered random forest (MLRF) framework

%Copyright (C) 2014 Alina Kuznetsova

%This program is free software: you can redistribute it and/or modify
%it under the terms of the GNU Lesser General Public License as published by
%the Free Software Foundation, either version 2.1 of the License, or
%(at your option) any later version.

%This program is distributed in the hope that it will be useful,
%but WITHOUT ANY WARRANTY; without even the implied warranty of
%MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%GNU Lesser General Public License for more details.

%You should have received a copy of the GNU General Public License
%along with this program.  If not, see <http://www.gnu.org/licenses/>.

%Author: Alina Kuznetsova

%% [Y,Ycl] = test(classifier,X,options)
%Input:
%   classifier - trained classifier
%   x - Nxdim feature vectors
%   options - various options
%Output:
%   Y - classes
%   Ycl - clusters, found on the first layer

function [Y,Ycl] = test(classifier,X,options)

    dim = size(X,2);
    N = size(X,1);

    if(~isfield(options,'verbose') || options.verbose)
        options.verbose = 1;
        disp('Testing...');
        disp(['Feature vector dimensions: ' num2str(dim)]);
        disp(['Number of samples: ' num2str(N)]);
        disp('Options: ');
        disp(options);
    end
    
    if (isfield(options.clusterOptions,'aggLevel'))
        Xagg = featureAggregation(X,options.clusterOptions.aggLevel);
    else
        Xagg = X;
    end
    
    if (isfield(classifier,'exclude'))
        X(:,classifier.exclude) = [];
        X = (X - repmat(classifier.Xmean,size(X,1),1))./repmat(classifier.Xstd,size(X,1),1); 
        
        if (isfield(options.clusterOptions,'aggLevel'))
             Xagg(:,classifier.XaggExclude) = [];
             Xagg = (Xagg - repmat(classifier.XaggMean,size(Xagg,1),1))./repmat(classifier.XaggStd,size(Xagg,1),1); 
        end
    end
    
    if (strcmp(classifier.type,'RF'))
        
        Y = forestTest(classifier.object, X);
        Ycl = ones(size(Y));
        
        %convert to symbols
        if (isfield(classifier,'code'))
            Yc = classifier.code(Y);
            
            if (size(Yc,1) ~= size(Y,1))
            	Yc = Yc';
            end
                   
            Y = Yc;
        end
        
    elseif (strcmp(classifier.type,'hierarchy'))
        %test 0-level classifier
        level0 = classifier.level0;
        
        clcode = classifier.clusters;
        
        [Ycl,Ysoft] = forestTest(level0.object, Xagg);
        
        Y = zeros(size(X,1),1);
        
        if (~options.clusteringOnly)
        
            %check 1-level classifiers
            for i=1:length(clcode)
                Xsub = X(Ycl==clcode(i),:);
                
                if (~isempty(Xsub))
                
                    if strcmp(classifier.level1(i).type,'RF')
                    
                    
                        %renormalize
                        if (isfield(classifier.level1(i),'exclude'))
                            Xsub(:,classifier.level1(i).exclude) = [];
                            Xsub = (Xsub - repmat(classifier.level1(i).Xmean,size(Xsub,1),1))./repmat(classifier.level1(i).Xstd,size(Xsub,1),1); 
                        end

                        Ysub = forestTest(classifier.level1(i).object, Xsub);
                    
                	elseif strcmp(classifier.level1(i).type,'const')
                    
                        Ysub = ones(size(Xsub,1),1)*classifier.level1(i).object;
                    end
            
                    Y(Ycl==clcode(i)) = Ysub;
                
                end
            end
        
            %convert to symbols
            if (isfield(classifier,'code'))
                Yc = classifier.code(Y);
                
                if (size(Yc,1) ~= size(Y,1))
                    Yc = Yc';
                end
                   
                Y = Yc;
            end 
        end
        
    else
        warning('Unknown classifier type');
    end

end
