function [model,sample2node] = treeTrain(X, Y, opts)
% Train a random tree
% X is NxD, each D-dimensional row is a data point
% Y is Nx1 discrete labels of classes
% returned model is to be directly plugged into treeTest

d= 5; % max depth of the tree

if nargin < 3, opts= struct; end
if isfield(opts, 'depth'), d= opts.depth; end

u= unique(Y);
u= sort(u);
[N, D]= size(X);
nd= 2^d - 1;
numInternals = (nd+1)/2 - 1;
numLeafs= (nd+1)/2;

%%choose the maximum number of parameters to store
    
weakModels = zeros(opts.mlength, numInternals); 

%this should be always possible if a sample is in memory: d is normally <= 100? :D
dataix = zeros(N,d-1);

if (nargout >1)
    sample2node = zeros(N,1);
end
    
leafdist= zeros(numLeafs, length(u)); % leaf distribution
leaftotal = zeros(numLeafs,1);

% Propagate data down the tree while training weak classifiers at each node
for n = 1: numInternals
    
    % get relevant data at this node
    if n==1 
        level = 0;
        reld = 1:size(dataix,1);
        Xrel= X;
        Yrel= Y;
    else
        level = floor(log2(n));
        reld = find(dataix(:, level)==n);
        Xrel = X(reld, :);
        Yrel = Y(reld);
    end
    
    

    if size(Xrel,1) >= opts.minspn
        % train weak model
        weakModels(:,n)= weakTrain(Xrel, Yrel, opts);
            
        
        % split data to child nodes
        yhat= weakTest(weakModels(:,n), Xrel, opts);
        
    else
            if(size(Xrel,1) == 0)
                weakModels(opts.mwIDidx,n)= 0;
            else
                weakModels(opts.mwIDidx,n) = opts.classifierID(1);
            end

        
        yhat = zeros(size(Xrel,1),1);
        
    end

    %weakModels{n}.classDist = histc(Yrel,u);
    
    % there is no split
    if (sum(yhat) <= opts.minscount)
        %to small amount of sample is separated
        yhat = zeros(size(yhat));
        weakModels(opts.nosplitidx,n) = 1; %go right (2n+1)
    elseif (length(yhat) - sum(yhat) <= opts.minscount)
        %to small amount of sample is separated
        yhat = ones(size(yhat));
        weakModels(opts.nosplitidx,n) = -1; %go left (2n)
    else
        weakModels(opts.nosplitidx,n) = 0; %there is a split
    end
    
    dataix(reld(yhat==1), level+1)= 2*n;
    dataix(reld(yhat==0), level+1)=2*n+1; % since yhat is in {0,1} and double
end

% Go over leaf nodes and assign class statistics
for n= (nd+1)/2 : nd
    reld= (dataix(:, d-1) == n);
    
    if (nargout>1)
        sample2node(reld==1) = n; 
    end
    
    hc = histc(Y(reld==1), u);
    %hc = hc + 1; % Dirichlet prior (no zero-size nodes)
    if (sum(hc)>0)
        leafdist(n - (nd+1)/2 + 1, :)= hc/sum(hc);
        leaftotal(n - (nd+1)/2 + 1) = sum(hc);
    end     
end

model.leafdist= leafdist;
model.leaftotal = leaftotal;
if(opts.saveIds)
    model.dataix = dataix;
end
model.depth= d;
model.classes= u;
model.weakModels= weakModels;
    
end
